#
#  $Id$
#
#  (C) Copyright 2003 Kai Sterker <kaisterker@linuxgames.com>
#  Part of the Adonthell Project http://adonthell.linuxgames.com
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY.
#
#  See the COPYING file for more details
#

import table
from damage import physical

#
# This extends the builtin property class, so that the get and set
# methods recieve the attribute to access as a string. That way, two
# generic get and set methods will be enough to care for all attributes.
#
class attr_property (property):
    def __init__ (self, attr, fget = None, fset = None):
        self.__attr = attr
        self.__get = fget
        self.__set = fset
        
    def __get__ (self, inst, type = None):
        if inst is None:
             return self
        if self.__get is None:
            raise AttributeError, "unreadable attribute"
        return self.__get (inst, self.__attr)

    def __set__ (self, inst, value):
        if self.__set is None:
            raise AttributeError, "'%s' is readonly" % self.__attr
        return self.__set (inst, self.__attr, value)

#
# This is the base class for all characters. It implements attributes
# and methods common to all characters. Characters that need additional
# attributes will have to extend this class.
#
class base (object):
    def __init__ (self, race):
        # -- the character's race (a string)
        self.__Race = race
        
        # -- the character's level
        self.__Level = 1
        
        # -- the character's experience
        self.__Experience = 0
        
        # -- storage for variable attributes the character has
        self.__Attributes = {}
        
        # -- modifiers to the maximum value of attributes
        self.__Modifiers = {"Strength":0, "Hardness":0, "Agility":0, \
            "Willpower":0, "Health":0, "Power":0, "Physical Resistance":0, \
            "Poison Resistance":0, "Fire Resistance":0, "Mental Resistance":0, \
            "Melee Attack":0, "Parry":0}
            
        # -- initialize attributes
        self.Strength  = table.fixed.initial (race, "Strength")
        self.Hardness  = table.fixed.initial (race, "Hardness")
        self.Agility   = table.fixed.initial (race, "Agility")
        self.Willpower = table.fixed.initial (race, "Willpower")
        self.Health    = table.variable.maximum (self, "Health")
        self.Power     = table.variable.maximum (self, "Power")

        
    # -- safely increase an attribute
    def increase (self, attribute, amount):
        value = self.__Attributes.get (attribute, 0) + amount
        try:
            maximum = getattr (self, "Max%s" % attribute)
        except AttributeError:
            print "*** warning: character.base.increase: character has no attribute called '%s'" % attribute
            return 0
            
        if value > maximum: 
            self.__Attributes[attribute] = maximum
            return 0
        else:
            self.__Attributes[attribute] = value
            return 1
        
    # -- safely decrease an attribute
    def decrease (self, attribute, amount):
        value = self.__Attributes.get (attribute, 0) - amount
           
        if value < 0: 
            self.__Attributes[attribute] = 0
            return 0
        else:
            self.__Attributes[attribute] = value
            return 1

    # -- add a modifier to the character
    def add_modifier (self, attribute, value):
        self.__Modifiers[attribute] += value

    # -- remove a modifier from the character
    def remove_modifier (self, attribute, value):
        self.__Modifiers[attribute] -= value
    
    # -- method to increase the character's experience
    #    returns how many levels the character gained
    def increase_experience (self, amount):
        # -- experience can only grow!
        if amount < 0: return 0
     
        self.__Experience += amount
        level = self.__Level
        
        # -- level up. if neccessary
        while self.__Experience >= 100 * self.__Level * (self.__Level + 1) / 2:
            self.__Level += 1
        
        # -- return number of levels the character gained
        return self.__Level - level
        
    # -- The get method returns fixed and variable attributes alike
    def get (self, attribute):
        return self.__Attributes[attribute]

    # -- The set method sets the current value of an attribute
    def set (self, attribute, value):
        self.__Attributes[attribute] = 0
        self.increase (attribute, value)
    
    # -- get the maximum of a fixed attribute (for internal use only)
    def get_max_fix (self, attribute):
        return table.fixed.maximum (self.__Race, attribute) + self.__Modifiers[attribute]
    
    # -- get the maximum of a derived attribute (for internal use only)
    def get_max_var (self, attribute):
        return table.variable.maximum (self, attribute) + self.__Modifiers[attribute]
    
    # -- return the fighting feat the character uses
    def get_feat (self): return None
    
    # -- return the weapon the character has equipped
    def get_weapon (self): return None
    
    # -- damage method for unarmed melee combat 
    def get_effects (self):
        d_range = table.fixed.get_melee (self.__Race)
        return [physical.physical (\
            (d_range[0] + self.__Attributes["Strength"],\
             d_range[1] + self.__Attributes["Strength"]))]

    # -- Current Stats (Read/Write)
    Strength        = attr_property ("Strength",    get, set)
    Hardness        = attr_property ("Hardness",    get, set)
    Agility         = attr_property ("Agility",     get, set)
    Willpower       = attr_property ("Willpower",   get, set)
    Health          = attr_property ("Health",      get, set)
    Power           = attr_property ("Power",       get, set)
    
    # -- Derived Stats (Read only, always at maximum)
    PhysicalResistance = attr_property ("Physical Resistance", get_max_var)
    PoisonResistance   = attr_property ("Poison Resistance",   get_max_var)
    FireResistance     = attr_property ("Fire Resistance",     get_max_var)
    MentalResistance   = attr_property ("Mental Resistance",   get_max_var)
    MeleeAttack        = attr_property ("Melee Attack",        get_max_var)
    Parry              = attr_property ("Parry",               get_max_var)
    
    # -- Allowed Maximum of Stats (Readonly)
    MaxStrength     = attr_property ("Strength",    get_max_fix)
    MaxHardness     = attr_property ("Hardness",    get_max_fix)
    MaxAgility      = attr_property ("Agility",     get_max_fix)
    MaxWillpower    = attr_property ("Willpower",   get_max_fix)
    MaxHealth       = attr_property ("Health",      get_max_var)
    MaxPower        = attr_property ("Power",       get_max_var)
    
    # -- Other attributes (Readonly)  
    Race            = property (lambda (self): self.__Race)
    Level           = property (lambda (self): self.__Level)
    Experience      = property (lambda (self): self.__Experience)
