#
#  $Id$
#
#  (C) Copyright 2003 Kai Sterker <kaisterker@linuxgames.com>
#  Part of the Adonthell Project http://adonthell.linuxgames.com
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY.
#
#  See the COPYING file for more details
#

import adonthell

#
# Helper class for storing fixed character attributes. One such class
# is instanciated for each race, and stored in fixed.__Table with the
# race as key.
#
class fixed_data (dict):
    def __init__ (self, file):
        self.get_state (file)
    
    # -- load fixed character data from file
    def get_state (self, file):
        
        # -- Read cost, minimum and maximum of each attribute
        while file.get_bool ():
            attribute = file.get_string ()
            cost = file.get_uint16 ()
            maximum = file.get_uint16 ()
            initial = file.get_uint16 ()
            
            # -- store them in the dictionary
            dict.__setitem__ (self, attribute, (cost, maximum, initial))
    
        # -- Read melee damage for unarmed combat
        minimum = file.get_uint16 ()
        maximum = file.get_uint16 ()
        
        self.MeleeDamage = (minimum, maximum)
        
#
# This class implements a global storage for character related
# constants. Note that it is static and should not be instanciated.
# (For that reason, it requires Python 2.2 or later to work)
#
class fixed (object):
    # -- dictionary storing the fixed data with the 
    #    race as key. The fixed data itself is a dictionary
    #    with attribute as key and a tuple containing cost to raise
    #    the attribute as well a maximum and initial rank.
    __Table = {}
    
    
    # -- speaking names for attribute-tuple access
    COST    = 0
    MAXIMUM = 1
    INITIAL = 2
    
    
    # -- Return the attributes available to a given race
    def attributes (race):
        return fixed.__Table.get (race, {}).keys ()
        
    
    # -- return the unarmed melee damage of given race    
    def get_melee (race):
        try:
            data = fixed.__Table[race]
            return data.MeleeDamage
        except:
            print "*** warning: character.table.fixed: race '%s' not in table" % str (race) 
            return (0, 1)
            
                
    # -- return the maximum rank the given race may gain 
    #    in the given attribute
    def maximum (race, attribute):
        return fixed.lookup (race, attribute, fixed.MAXIMUM, 0)
        
        
    # -- return the cost to advance the given attribute 
    #    of the given race
    def cost (race, attribute):
        return fixed.lookup (race, attribute, fixed.COST, 1000)


    # -- return initial value of the given attribute
    def initial (race, attribute):
        return fixed.lookup (race, attribute, fixed.INITIAL, 0)
        
    
    # -- actual lookup operation
    def lookup (race, attribute, key, errval):
        try:
            dict = fixed.__Table[race]
        except:
            print "*** warning: character.table.fixed: race '%s' not in table" % str (race) 
            return errval
            
        try:
            return dict[attribute][key]
        except:
            print "*** warning: character.table.fixed: attribute '%s' not in table of race '%s'" % (attribute, race) 
            return errval    
    
    
    # -- load fixed table from disk
    def get_state ():
        # -- try to open file
        file = adonthell.igzstream ()
        file.open ("advancement.data")
        
        if not file.is_open ():
            print "*** error: character.table.fixed: cannot open 'advancement.data'"
            return
        
        # -- as long as 'records' are in file, read them
        while file.get_bool ():
            # -- get name of race
            race = file.get_string ()
            
            # -- get fixed data for that race
            fixed.__Table[race] = fixed_data (file)
    
        # -- cleanup
        file.close ()
    
    
    # -- make all methods static
    attributes = staticmethod (attributes)
    get_melee = staticmethod (get_melee)
    get_state = staticmethod (get_state)
    maximum = staticmethod (maximum)
    initial = staticmethod (initial)
    lookup = staticmethod (lookup)
    cost = staticmethod (cost)
    

# Load the fixed data from disk. Happens when this module
# is imported for the first time
fixed.get_state ()


#
# This class contains formulas to calculate the maximum of
# certain attributes from other attributes at runtime.
#
class variable (object):
    __Table = {\
        "Melee Attack":         lambda (char): 10 * char.Strength,\
        "Carrying Capacity":    lambda (char): 10 * char.Strength, \
        "Health":               lambda (char): 10 * char.Hardness, \
        "Physical Resistance":  lambda (char): 1 * char.Hardness, \
        "Poison Resistance":    lambda (char): 1 * char.Hardness, \
        "Fire Resistance":      lambda (char): 1 * char.Hardness, \
        "Ranged Attack":        lambda (char): 10 * char.Agility, \
        "Dodge":                lambda (char): 10 * char.Agility, \
        "Parry":                lambda (char): 10 * char.Agility, \
        "Power":                lambda (char): 10 * char.Willpower, \
        "Mental resistance":    lambda (char): 1 * char.Willpower, \
    }
    
    def maximum (character, attribute):
        return variable.__Table[attribute](character)
        
    maximum = staticmethod (maximum)
